#include "main.h"

LIST compile_opt  = { NULL, NULL, 0 };      // option for COMPILE
LIST assemble_opt = { NULL, NULL, 0 };      // option for ASSEMBLE
LIST link_opt     = { NULL, NULL, 0 };      // option for LINK

LIST cpp_files   = { NULL, NULL, 0 };       // list files for COMPILE
LIST asm_files   = { NULL, NULL, 0 };       // list files for ASSEMBLE
LIST obj_files   = { NULL, NULL, 0 };       // list all .OBJ files
LIST lib_files   = { NULL, NULL, 0 };       // list all .LIB files

// == TRUE, if need ...
int _compile = TRUE;            // default == TRUE
int _assemble = TRUE;           // default == TRUE
int _link = TRUE;               // default == TRUE

int _debug = FALSE;             // is DEBUG 
int _nodefault = FALSE;		// is no default .lib & .obj
int _keep_temp = FALSE;		// save CL386.$$$, TASM.$$$ & TLINK.$$$
int _map_file = FALSE;		// generate .MAP file

static void delete_all( void );
static void delete_list( void );

void main( int argc, char *argv[] ) {

    if ( argc == 1 )
        help();

    atexit( delete_all );		// On exit delete all ...

    get_config_file( argv[0] );		// Search & read CL386.CFG file

    get_environ_var( "CL386" );		// read env var CL386 

    get_params( argc - 1, argv + 1 );	// read args

    if ( cpp_files.count == 0 && asm_files.count == 0 && \
         obj_files.count == 0 && lib_files.count == 0 )	//  No files - help !
         help();

    compile_response();		// Create CL386.$$$
    assemble_response();	// Create TASM.$$$
    link_response();		// Create TLINK.$$$

    if ( _compile && cpp_files.count == 0 )	// No files for compile
        _compile = 0;
    if ( _assemble && asm_files.count == 0 )	// No files for assemble
        _assemble = 0;
    if ( _link && obj_files.count == 0 && lib_files.count == 0 )
        _link = 0;				// No files for link 

    delete_list();	// free max memory

    compile();		// 
    assemble();		//
    link();		//
	
    delete_response();	// delete CL386.$$$, TASM.$$$ & TLINK.$$$ if need
}

void delete_list( void ) {	// delete all list from memory
    LIST_done( &compile_opt );
    LIST_done( &assemble_opt );
    LIST_done( &link_opt );
    LIST_done( &cpp_files );
    LIST_done( &asm_files );
    LIST_done( &obj_files );
    LIST_done( &lib_files );
}

void delete_all( void ) {	// for atexit()
    delete_list();
    delete_response();
}