#include "main.h"

// Usage
static char usage[] = \
"CL386 Version " VERSION " (c) Kirill Joss. Compile " __DATE__ "\n"
"Usage: CL386.EXE [options] files\n"
"Options:\n"
"/i[+|-]  dump preprocessed file   @name    specify response file\n"
"/l[+|-]  dump listing file        /w-xxx   disable a warning\n"
"/A[+|-]  disable extensions       /Cxxx    codegen parameters\n"
"/Dxxx    define something         /Enn     max number of errors\n"
"/Ipath   specify include path     /Oxxx    optimzer parameters\n"
"/e[+|-]  dump errors to file\n"
"------------------------------------------------------------------\n"
"/K       keep response files      /Lxxx    specify .LIB & .OBJ dir\n"
"/a       generate .ASM file       /c       generate .OBJ file\n"
"/v       full debug info          /m       generate .MAP file\n"
"/?, /h   this help                /n       no default .LIB & .OBJ\n";
//  "\n\t(c) 1997 Kirill Joss. Compiled "__DATE__ "\n";

void help( void ) {
    printf( usage );
    exit(0);
}

void get_option( char *param ) {
    switch( param[1] ) {

        case 'i' :
            if ( param[2] == '-' ) { 			// -i-
                LIST_add( &compile_opt, "-i" );
                return;
            }
            else if ( param[2] == '+' || param[2] == '\0' ) { // -i or -i+ :)
                LIST_add( &compile_opt, "+i" );
                return;
            }
            break;

        case 'e' :
            if ( param[2] == '-' ) { 			// -e-
                LIST_add( &compile_opt, "-e" );
                return;
            }
            else if ( param[2] == '+' || param[2] == '\0' ) { // -e or -e+ :)
                LIST_add( &compile_opt, "+e" );
                return;
            }
            break;

         case 'l' :
            if ( param[2] == '-' ) { // -i-
                LIST_add( &compile_opt, "-l" );
                return;
            }
            else if ( param[2] == '+' || param[2] == '\0' ) { // -i or -i+ :)
                LIST_add( &compile_opt, "+l" );
                return;
            }
            break;

        case 'A' :
            if ( param[2] == '-' ) { // /A-
                LIST_add( &compile_opt, "-A" );
                return;
            }
            else if ( param[2] == '+' || param[2] == '\0' ) { // -A or -A+
                LIST_add( &compile_opt, "+A" );
                return;
            }
            break;

        case 'w' :
        case 'C' :
        case 'D' :
        case 'E' :
        case 'I' :
        case 'O' :
            LIST_add( &compile_opt, param );
            return;

        case '?' :
        case 'h' :
            help(); /* Not return ! */

        case 'm' :
            _map_file = TRUE;
        case 'L' :
            LIST_add( &link_opt, param );
            return;

        case 'n':
            _nodefault = TRUE;
            return;

        case 'K':
            _keep_temp = TRUE;
            return;

        case 'v' :
            _debug = TRUE;
            LIST_add( &assemble_opt, "/zi" );
            LIST_add( &link_opt, "/l" );
            LIST_add( &link_opt, "/s" );
            LIST_add( &link_opt, "/v" );
            return;

        case 'a' :
            _compile = TRUE;
            _assemble = _link = FALSE;
            return;

        case 'c' :
            _compile = _assemble = TRUE;
            _link = FALSE;
            return;

    }
    fprintf( stderr, "Warning! Parameter error : %s\n", param );
}
