#include "main.h"

void compile_response( void ) {
    if ( _compile && cpp_files.count > 0 ) {
        FILE *frsp;
        int i;

        frsp = fopen( COMPILE_RESPONSE, "wt" );

        if ( frsp == NULL )
            error( 1, COMPILE_RESPONSE );

#if defined( COMPILE_DEFAULT_OPTIONS )
        fprintf( frsp, "%s ", COMPILE_DEFAULT_OPTIONS );
#endif

        for ( i = 0; i < compile_opt.count; i++ )
            fprintf( frsp, " %s", LIST_item( &compile_opt, i) );

        fprintf( frsp, "\n" );

#if defined( COMPILE_DEFAULT_FILES )
        fprintf( frsp, "%s\n", COMPILE_DEFAULT_FILES );
#endif

        for ( i = 0; i < cpp_files.count; i++ )
            fprintf( frsp, "%s\n", LIST_item( &cpp_files, i) );

        fclose( frsp );
    }
}

void assemble_response(void) {
    if ( _assemble && asm_files.count > 0 ) {
        FILE *frsp;
        int j, i;

        frsp = fopen( ASSEMBLE_RESPONSE, "wt" );

        if ( frsp == NULL )
            error( 1, ASSEMBLE_RESPONSE );

        for ( j = 0; j < asm_files.count; j++ ) {

#if defined( ASSEMBLE_DEFAULT_OPTIONS )
            fprintf( frsp, "%s ", ASSEMBLE_DEFAULT_OPTIONS );
#endif

            for ( i = 0; i < assemble_opt.count; i++ )
                fprintf( frsp, " %s", LIST_item( &assemble_opt, i) );

            fprintf( frsp, " %s;\n", LIST_item( &asm_files, j) );
        }
        fclose( frsp );
    }
}


static char name[ _MAX_FNAME ];

static void get_first( LIST *list ) {
    if ( list == NULL )
        error( 1, "Pointer is NULL!" );
    _splitpath( LIST_item( list, 0 ), NULL, NULL, name, NULL );
}

void link_response( void ) {
    if ( _link && ( obj_files.count > 0 || lib_files.count > 0) ) {
        FILE *frsp;
        int i;

        frsp = fopen( LINK_RESPONSE, "wt" );

        if ( frsp == NULL )
            error( 1, LINK_RESPONSE );

#if defined( LINK_DEFAULT_OPTIONS )
        fprintf( frsp, "%s ", LINK_DEFAULT_OPTIONS );
#endif

        for ( i = 0; i < link_opt.count; i++ )
            fprintf( frsp, "%s ", LIST_item( &link_opt, i) );

    if ( !_nodefault ) {
        if ( _debug )
#if defined( LINK_DEBUG_OBJ )
            fprintf( frsp, "%s ", LINK_DEBUG_OBJ )
#endif
            ;
        else
#if defined( LINK_DEFAULT_OBJ )
            fprintf( frsp, "%s ", LINK_DEFAULT_OBJ )
#endif
            ;
    }

        if ( obj_files.count > 0 )      // Get NAME .EXE file ...
            get_first( &obj_files );
        else
            get_first( &lib_files );

        for ( i = 0; i < obj_files.count; i++ )
            fprintf( frsp, "%s ", LIST_item( &obj_files, i) );

        fprintf( frsp, ",%s,%s,", name, ( _map_file ? name : "nul" ) );

    if ( !_nodefault ) {
    
#if defined( LINK_DEFAULT_LIB )
        fprintf( frsp, "%s ", LINK_DEFAULT_LIB );
#endif

    }
    
        for ( i = 0; i < lib_files.count; i++ )
            fprintf( frsp, "%s ", LIST_item( &lib_files, i) );

        fclose( frsp );
    }
}

void delete_response( void ) {
    if ( !_keep_temp ) {
        remove( COMPILE_RESPONSE );
        remove( ASSEMBLE_RESPONSE );
        remove( LINK_RESPONSE );
    }
}


void exec( char *program, char *response ) {
    static char *args[3];
    int err;

    args[0] = program;
    args[1] = response;
    args[2] = NULL;

    err = spawnvp( P_WAIT, program, args );

    if ( err <  0 )
        error( 1, program );
    else if ( err > 0 )
        error( 0, program );
}

void compile( void ) {
    if ( _compile ) {
        exec( COMPILE_PROGRAM, "/f" COMPILE_RESPONSE );
    }
}

void assemble( void ) {
    if ( _assemble ) {
        exec( ASSEMBLE_PROGRAM, "@" ASSEMBLE_RESPONSE );
    }
}

void link( void ) {
    if ( _link ) {
        exec( LINK_PROGRAM, "@" LINK_RESPONSE );
    }
}
